#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

template <typename scalar_t>
__global__ void unrolled_loop_kernel(
    const scalar_t* __restrict__ input,
    const scalar_t* __restrict__ weight,
    const scalar_t* __restrict__ bias,
    scalar_t* __restrict__ output,
    const int batch_size,
    const int in_features,
    const int out_features,
    const float subtract_value,
    const float multiply_value) {

    const int row = blockIdx.x;
    const int col = blockIdx.y;
    const int lane = threadIdx.x;

    scalar_t sum = 0;
    const int input_base = row * in_features;
    const int weight_base = col * in_features;

    // Manually unrolled loop with 4x unrolling
    #pragma unroll 4
    for (int k = lane; k < in_features; k += 32) {
        sum += input[input_base + k] * weight[weight_base + k];
    }

    // Improved warp reduction with unrolled steps
    unsigned mask = 0xffffffff;
    #pragma unroll
    for (int offset = 16; offset > 0; offset /= 2) {
        sum += __shfl_down_sync(mask, sum, offset);
    }

    if (lane == 0) {
        scalar_t result = sum + bias[col];
        result = (result - subtract_value) * multiply_value;
        output[row * out_features + col] = result > 0 ? result : 0;
    }
}

torch::Tensor forward(
    torch::Tensor input,
    torch::Tensor weight,
    torch::Tensor bias,
    float subtract_value,
    float multiply_value) {

    const int batch_size = input.size(0);
    const int in_features = input.size(1);
    const int out_features = weight.size(0);

    auto output = torch::empty({batch_size, out_features}, input.options());

    dim3 blocks(batch_size, out_features);
    dim3 threads(32);

    AT_DISPATCH_FLOATING_TYPES(input.scalar_type(), "unrolled_loop_kernel", ([&] {
        unrolled_loop_kernel<scalar_t><<<blocks, threads>>>(
            input.data_ptr<scalar_t>(),
            weight.data_ptr<scalar_t>(),
            bias.data_ptr<scalar_t>(),
            output.data_ptr<scalar_t>(),
            batch_size,
            in_features,
            out_features,
            subtract_value,
            multiply_value
        );
    }));

    return output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Matmul with fused ops using unrolled loops");
}